#include "ofApp.h"
#include "myTrueTypeFont.h"

#include "Poco/RegularExpression.h"
using Poco::RegularExpression;

#define NUM_ERASURE_DEPTHS 6
#define NUM_SEQUENCES_PER_GROUP 3

// BASIC METHODS //
//--------------------------------------------------------------
void ofApp::setup(){
    
    // DRAWING BASICS
    
    ofSetBackgroundAuto(true);
    ofSetVerticalSync(true);
    ofEnableAlphaBlending();
    ofBackground(255);
    globalAlphaFactor = 1;
    
    
    // LAYOUT
    // ** set up layoutVec.  First number is layout number bool, second is fontsize
    
    fontSizeVec.push_back(44); // index = 0 | large font small margin
    fontSizeVec.push_back(28); // index = 1 | ebooklayout 1
    fontSizeVec.push_back(34); // index = 2 | synthesis of the two
    layoutIndex = 2;
    isJustified = true;
    
    myFont = std::shared_ptr<myTrueTypeFont>(new myTrueTypeFont(false, fontSizeVec[layoutIndex]));
    myHeadingFont = std::shared_ptr<myTrueTypeFont>(new myTrueTypeFont(true, 25));
    
    
    // ACCELEROMETER
    
    ofxAccelerometer.setup();
    hasCompass = coreLocation.startHeading();
    
    
    // CREATE OUR TEXT PARTICLES
    
    createParticlesFromXml();
    
    
    // BOOLS
    
    indeterminate = false;
    sequenceGroupChange = false;
    subroutine = false;
    resetRoutine = true;
    
    subroutineSequenceIndex = 0;
    masterRoutineSequenceIndex = 0;
    masterRoutineIterator = 0;
    subroutineIterator = 0;
    
    
    // DEBUG TOOLS
    
    simulator = false;
    slowFrameRate = false;
    displayStats = false;
    enableSlowFrameRate = false;
    
    
    // DEFINE SEQUENCE GROUPS
    
    ofVec2f groupOne(0,2);
    ofVec2f groupTwo(3,5);
    ofVec2f groupThree(6,8);
    sequenceGroups.push_back(groupOne);
    sequenceGroups.push_back(groupTwo);
    sequenceGroups.push_back(groupThree);
    
    
    // INITIALIZE SEQUENCE GROUP
    
    sequenceGroupIndex = 0;
    sequenceGroupLowerBound = sequenceGroups[sequenceGroupIndex].x;
    sequenceGroupUpperBound = sequenceGroups[sequenceGroupIndex].y;
    
    sequenceIndex = sequenceGroupLowerBound;
    
    prevSequenceIndex.resize(6);
    for (int i = 0; i < prevSequenceIndex.size(); i++)
    {
        prevSequenceIndex[i] = sequenceGroupLowerBound;
    }
    
}
//--------------------------------------------------------------
void ofApp::update(){
    
    // GET & CORRECT X AXIS ORIENTATION (*ERASURE DEPTH*)
    
    accOrientCheck = ofxAccelerometer.getOrientation();
    temp = accOrientCheck.x;
    accOrientCheck.x -=90;
    accOrientCheck.x *=-1;
    accOrientCorrected.x = accOrientCheck.x;
    
    
    // GET & CORRECT HEADING (SEQUENCE GROUP)
    
    headingCheck = ofLerpDegrees(headingCheck, -coreLocation.getTrueHeading(), 0.7);
    
    if (loopCounter < 5)
    {
        headingDisplay = 1.5;
    }
    else if (loopCounter==5)
    {
        headingDiff = headingCheck - 30;
        
    }
    else if (loopCounter > 5)
    {
        headingCorrected = headingCheck - headingDiff;
        headingCorrected = (int)headingCorrected % 360;
        float headingDecimals = headingCheck - floor(headingCheck);
        headingCorrected += headingDecimals;
        if (headingCorrected < 0)
        {
            headingCorrected += 360;
        }
        
        headingDisplay = ofMap(headingCorrected, 0, 360, 0.000, 6.000, true);
        
    }
    
    loopCounter++;
    
    // SET xAxisProbabilityTarget and yAxis
    
    xAxisProbabilityTarget = ofMap(accOrientCorrected.x, 90, 250, 0, 280, true);
    yAxis = ofMap(accOrientCorrected.y, 0, 180, 1500, 10000, true);
    
    // SET ERASURE DEPTH DISPLAY BASED ON X AXIS PROB TARGET
    
    // set bounds
    int lowerBound1 = 0;    // erasure depth 0 (does note exist)
    int lowerBound2 = 34;   // erasure depth 1
    int lowerBound3 = 86;  // erasure depth 2
    int lowerBound4 = 135;  // erasure depth 3
    int lowerBound5 = 185;  // erasure depth 4
    int lowerBound6 = 233;  // erasure depth 5
    int lowerBound7 = 290;  // erasure depth 6 (does not exist)
    
    
    if ((xAxisProbabilityTarget >= lowerBound1) && (xAxisProbabilityTarget <= lowerBound2))
    {
        erasureDepthDisplay = ofMap(xAxisProbabilityTarget, lowerBound1, lowerBound2, 0, 0.999, true);
        erasureDepthCurrent = 0;
    }
    else if ((xAxisProbabilityTarget > lowerBound2) && (xAxisProbabilityTarget <= lowerBound3))
    {
        erasureDepthDisplay = ofMap(xAxisProbabilityTarget, lowerBound2, lowerBound3, 1, 1.999, true);
        erasureDepthCurrent = 1;
    }
    else if ((xAxisProbabilityTarget > lowerBound3) && (xAxisProbabilityTarget <= lowerBound4))
    {
        erasureDepthDisplay = ofMap(xAxisProbabilityTarget, lowerBound3, lowerBound4, 2, 2.999, true);
        erasureDepthCurrent = 2;
    }
    else if ((xAxisProbabilityTarget > lowerBound4) && (xAxisProbabilityTarget <= lowerBound5))
    {
        erasureDepthDisplay = ofMap(xAxisProbabilityTarget, lowerBound4, lowerBound5, 3, 3.999, true);
        erasureDepthCurrent = 3;
    }
    else if ((xAxisProbabilityTarget > lowerBound5) && (xAxisProbabilityTarget <= lowerBound6))
    {
        erasureDepthDisplay = ofMap(xAxisProbabilityTarget, lowerBound5, lowerBound6, 4, 4.999, true);
        erasureDepthCurrent = 4;
    }
    else if ((xAxisProbabilityTarget > lowerBound6) && (xAxisProbabilityTarget <= lowerBound7))
    {
        erasureDepthDisplay = ofMap(xAxisProbabilityTarget, lowerBound6, lowerBound7, 5, 5.5);
        erasureDepthCurrent = 5;
    }
    
    
    // MANAGE *SEQUENCE GROUP* BASED ON HEADING DISPLAY
    
    if ((headingDisplay >= 0) && (headingDisplay < 1))
    {
        sequenceGroupIndex = 0;
    }
    else if ((headingDisplay >=1) && (headingDisplay < 2))
    {
        sequenceGroupIndex = 1;
    }
    else if ((headingDisplay >=2) && (headingDisplay < 3))
    {
        sequenceGroupIndex = 2;
    }
    else if ((headingDisplay >= 3) && (headingDisplay < 4))
    {
        sequenceGroupIndex = 0;
    }
    else if ((headingDisplay >= 4) && (headingDisplay < 5))
    {
        sequenceGroupIndex = 1;
    }
    else if ((headingDisplay >=5) && (headingDisplay < 6))
    {
        sequenceGroupIndex = 2;
    }
    
    
    // CHECK FOR CHANGE IN SEQUENCE GROUP
    
    if (sequenceGroupIndex != prevSequenceGroupIndex)
    {
        
        sequenceGroupChange = true;
        resetRoutine = true;
        sequenceGroupLowerBound = sequenceGroups[sequenceGroupIndex].x;
        sequenceGroupUpperBound = sequenceGroups[sequenceGroupIndex].y;
        
        sequenceIndex = sequenceGroupLowerBound;
        
        prevSequenceIndex.resize(6);
        for (int i = 0; i < prevSequenceIndex.size(); i++)
        {
            prevSequenceIndex[i] = sequenceGroupLowerBound;
        }
    }
    
    
    // CHECK FOR CHANGE IN DEVICE POSITION & CHANGE SEQUENCE INDEX ACCORDINGLY
    
    if (erasureDepthCurrent != prevErasureDepth)
    {
        resetRoutine = true;
        sequenceIndex = prevSequenceIndex[erasureDepthCurrent] + 1;
    }
    
    // CREATE "NOISE" AT HIGHER ERASURE LEVELS
    
    if (erasureDepthCurrent > 3)
    {
        // SET SUBTEXT COORDINATES
        poemCoordinates.x = sequenceGroupIndex;
        poemCoordinates.y = erasureDepthCurrent;
        
        // SAMPLE VALUES
        routineLengthRange.x = 5;
        routineLengthRange.y = 8;
        masterRoutineLengthRange.x = 3;
        masterRoutineLengthRange.y = 5;
        subroutineLengthRange.x = 2;
        subroutineLengthRange.y = 4;
        
        if (prevPoemCoordinates!=poemCoordinates)
        {
            resetRoutine = true;
        }
        
        if (resetRoutine)
        {
            float temp = 0;
            int random = 0;
            float random2;
            resetRoutine = false;
            subroutine = false;
            routineIterator = 0;
            subroutineIterator = 0;
            masterRoutineIterator = 0;
            
            // DETERMINE MASTER AND SUBROUTINE SEQUENCE INDEXES
            
            // %50 of the time, don't make the masterRoutineSequenceIndex default
            // to the sequenceGroupLowerBound
            random2 = ofRandom(2);
            if (random2 > 1)
            {
                masterRoutineSequenceIndex = sequenceGroupLowerBound;
                temp = NUM_SEQUENCES_PER_GROUP - 1.01;
                random = floor(ofRandom(temp)) + 1;
                subroutineSequenceIndex = sequenceIndex + random;
            }
            else
            {
                int temp2 = floor(ofRandom(2.99));
                masterRoutineSequenceIndex = sequenceGroupLowerBound + temp2;
                subroutineSequenceIndex = masterRoutineSequenceIndex - 1;
                if (subroutineSequenceIndex < sequenceGroupLowerBound)
                {
                    subroutineSequenceIndex = sequenceGroupUpperBound;
                }
            }
            
            routineLength = (int)ofRandom(routineLengthRange.x, routineLengthRange.y);
            subroutineLength = (int)ofRandom(subroutineLengthRange.x, subroutineLengthRange.y);
            masterRoutineLength = (int)ofRandom(masterRoutineLengthRange.x, masterRoutineLengthRange.y);
            
        }
        
        if (!subroutine)
        {
            // SET SEQUENCE INDEX TO MASTER
            sequenceIndex = masterRoutineSequenceIndex;
            
            masterRoutineIterator++;
            if (masterRoutineIterator==masterRoutineLength)
            {
                subroutine = true;
            }
        }
        
        if (subroutine)
        {
            // SET SUBROUTINE INDEX TO SUBROUTINE
            sequenceIndex = subroutineSequenceIndex;
            
            subroutineIterator++;
            if (subroutineIterator==subroutineLength)
            {
                subroutine = false;
                subroutineIterator = 0;
                subroutineLength = (int)ofRandom(subroutineLengthRange.x, subroutineLengthRange.y);
                masterRoutineLength = (int)ofRandom(masterRoutineLengthRange.x, masterRoutineLengthRange.y);
            }
        }
        
        if (routineIterator==routineLength)
        {
            resetRoutine = true;
        }
        
        routineIterator++;
        prevPoemCoordinates = poemCoordinates;
    }
    
    
    // UPDATE OUR PARTICLES
    
    // MAKE SURE WE AREN'T SENDING AN OUT OF RANGE SEQUENCE INDEX
    if (sequenceIndex > sequenceGroupUpperBound)
    {
        sequenceIndex = sequenceGroupLowerBound;
    }
    
    bool firstWord = true;
    ofVec4f prevPrevWordInfo;
    // (STORES VITAL STATS FROM PREV WORD)
    //      .x = currentPosition.x
    //      .y = currentPosition.y
    //      .z = wordId
    //      .w = erasureDepth
    
    for (int i=0; i < textParticleVec.size(); i++)
    {
        if (firstWord)
        {
            ofVec4f firstWordInfo(-1,-1,-1,-1); // a null marker, since there is no previous word for the first word we draw
            prevPrevWordInfo = textParticleVec[i]->update(mouseMovement, erasureDepthCurrent, xAxisProbabilityTarget, indeterminate, yAxis, sequenceIndex, firstWordInfo);
            firstWord = false;
        }
        else
        {
            ofVec4f prevWordInfo = prevPrevWordInfo;
            prevPrevWordInfo = textParticleVec[i]->update(mouseMovement, erasureDepthCurrent, xAxisProbabilityTarget, indeterminate, yAxis, sequenceIndex, prevWordInfo);
        }
    }
    
    
    // RESET PREVIOUS STATE VARS
    
    prevSequenceIndex[devicePosition.x] = sequenceIndex;
    prevErasureDepth = erasureDepthCurrent;
    prevSequenceGroupIndex = sequenceGroupIndex;
    
}
//--------------------------------------------------------------
void ofApp::draw(){
    
    float orientX = erasureDepthDisplay;
    int orientY = (int)accOrientCorrected.y;
    string headingString = ofToString(headingDisplay, 2);
    string orientXString = ofToString(orientX, 2);
    string orientYString = ofToString(orientY);
    string slash = "/";
    
    float yPos = 1957;
    float orientXStringPos = 700;
    float orientXStringPosFinal;
    float slash1 = 765;
    float headingStringPos = 805;
    float headingStringPosFinal;
    
    float orientXStringWidth = myHeadingFont->stringWidth(orientXString);
    float orientYStringWidth = myHeadingFont->stringWidth(orientYString);
    float headingStringWidth = myHeadingFont->stringWidth(headingString);
    
    // PERFECTLY ALIGN ORIENT X
    if (orientXStringWidth < 22) // one digit, which is around 16 pixels wide
    {
        orientXStringPosFinal = orientXStringPos - 8;
    }
    else if ((orientXStringWidth > 22) && (orientXStringWidth < 40)) // two digits, around 30 pixels wide
    {
        orientXStringPosFinal = orientXStringPos - 15;
    }
    else if (orientXStringWidth > 40) // three digits, around 50 pixels
    {
        orientXStringPosFinal = orientXStringPos - 25;
    }
    
    
    // PERFECTLY ALIGN HEADING
    if (headingStringWidth < 22) // one digit, which is around 16 pixels wide
    {
        headingStringPosFinal = headingStringPos - 8;
    }
    else if ((headingStringWidth > 22) && (headingStringWidth < 40)) // two digits, around 30 pixels wide
    {
        headingStringPosFinal = headingStringPos - 15;
    }
    else if (headingStringWidth > 40) // three digits, around 50 pixels
    {
        headingStringPosFinal = headingStringPos - 25;
    }
    
    
    myHeadingFont->drawString(headingString, orientXStringPosFinal, yPos);
    myHeadingFont->drawString(slash, slash1, yPos);
    myHeadingFont->drawString(orientXString, headingStringPos, yPos);
    
    
    // BIND
    ofSetColor(0);
    myFont->bind();
    for (int i=0; i < textParticleVec.size(); i++)
    {
        textParticleVec[i]->draw(yAxis, motionDetected, globalAlphaFactor, sequenceGroupChange);
    }
    myFont->unbind();
    
    if (displayStats)
    {
        std::string infoString = makeInfoString();
        ofPushMatrix();
        ofScale(2, 2);
        ofDrawBitmapStringHighlight(infoString, 35, 200);
        ofPopMatrix();
    }
    
    
    if (myMousePressedSingleLoop) myMousePressedSingleLoop = false;
    if (sequenceGroupChange)
    {
        sequenceGroupChange = false;
        cout << "sequenceGroupChange is FALSE" << endl;
    }
    
}